#ifndef _AxeIntrinsic_h_
#define _AxeIntrinsic_h_

typedef int AxeIterationType;

#define AxeIsModified      1
#define AxeSafeClose       2
#define AxeForceClose      3
#define AxeSave            4
#define AxeSaveAndClose    5

#define noname             "No name"

#define AXE_SERVER         "AXE_SERVER"
#define AXE_COAXE          "AXE_COAXE"
#define AXE_FAXE           "AXE_FAXE"
#define AXE_POLEAXE        "AXE_POLEAXE"

#endif /* _AxeIntrinsic_h_ */

