/* String Constants. */
#define INFO_FILE_LABEL                 "File:"
#define INFO_NODE_LABEL                 "Node:"
#define INFO_PREV_LABEL                 "Prev:"
#define INFO_ALTPREV_LABEL              "Previous:"
#define INFO_NEXT_LABEL                 "Next:"
#define INFO_UP_LABEL                   "Up:"
#define INFO_MENU_LABEL                 "\n* Menu:"
#define INFO_MENU_ENTRY_LABEL           "\n* "
#define INFO_XREF_LABEL                 "*Note"
#define TAGS_TABLE_END_LABEL            "\nEnd Tag Table"
#define TAGS_TABLE_BEG_LABEL            "Tag Table:\n"
#define INDIRECT_TAGS_TABLE_LABEL       "Indirect:\n"
#define TAGS_TABLE_IS_INDIRECT_LABEL    "(Indirect)"

/* Character Constants. */
#define INFO_COOKIE '\037'
#define INFO_FF     '\014'
#define INFO_TAGSEP '\177'

#define INFO_CHAPTER                    "\n*"
#define INFO_SECTION                    "\n="
#define INFO_SUB_SECTION                "\n-"
#define INFO_SUB_SUB_SECTION            "\n."

#define IS_NODE           1
#define IS_TAG_TABLE      2
#define IS_INDIRECT       3
#define IS_INDIRECT_LABEL 4
#define IS_TAG            5
#define IS_NEXT           6
#define IS_PREV           7
#define IS_UP             8
#define IS_MENU           9
#define IS_ENTRY         10
#define IS_XREF          11

#define IS_SECTIONING    12
